"""XCAP module. Provides XCAP application abstractions and an XCAP client"""

import urllib2
import re

import urllib
import httplib
import base64

from urlparse import urlparse
from lxml import etree


class HTTPRequest(urllib2.Request):
    """Hack urllib2.Request to support PUT and DELETE methods."""
    
    def __init__(self, url, method="GET", data=None, headers={},
                 origin_req_host=None, unverifiable=False):
        urllib2.Request.__init__(self,url,data,headers,origin_req_host,unverifiable)
        self.method = method
    
    def get_method(self):
        return self.method
    

class HTTPResponse(object):
    
    def __init__(self, code, headers, body=None):
        self.code = code
        self.headers = headers
        self.body = body
    
    def get_header(self, header):
        return self.headers.get(header)

    def __repr__(self):
        if self.body is None:
            bodylen = None
        else:
            bodylen = len(self.body)
        return "<%s.%s code=%d, datalen=%s>" % (self.__module__, self.__class__.__name__, self.code, bodylen)


class XCAPClient(object):

    def __init__(self, xcap_root, auth=None):
        self.xcap_root = xcap_root
        self.auth = auth

    def _execute_request(self, method, url, user, realm, password, headers={}, data=None):
        authhandlers = []
        if self.auth == "basic":
            authhandlers.append(urllib2.HTTPBasicAuthHandler())
        elif self.auth == "digest":
            authhandlers.append(urllib2.HTTPDigestAuthHandler())
        else:
            authhandlers.append(urllib2.HTTPBasicAuthHandler())
            authhandlers.append(urllib2.HTTPDigestAuthHandler())
        for handler in authhandlers:
            handler.add_password(realm, url, user, password)
        opener = urllib2.build_opener(*authhandlers)
        req = HTTPRequest(url, method=method, headers=headers, data=data)
        try:
            r = opener.open(req)
            data = r.read()
            r.fp._sock.recv = None
            return HTTPResponse(r.code, r.headers, data)
        except urllib2.HTTPError, e:
            headers = getattr(e, 'headers', {})
            return HTTPResponse(e.code, headers)

    def get_resource(self, account, password, application, docname, node=None, headers={}):
        username, domain = account.split('@', 1)
        uri = "%s/%s/users/%s/%s" % (self.xcap_root, application, account, docname)
        if node:
            uri += '~~' + node
        return self._execute_request("GET", uri, username, domain, password, headers)

    def put_resource(self, account, password, application, docname, resource, node=None, headers={}):
        username, domain = account.split('@', 1)
        uri = "%s/%s/users/%s/%s" % (self.xcap_root, application, account, docname)
        if node:
            uri += '~~' + node
        return self._execute_request("PUT", uri, username, domain, password, headers, resource)

    def delete_resource(self, account, password, application, docname, node=None, headers={}):
        username, domain = account.split('@', 1)
        uri = "%s/%s/users/%s/%s" % (self.xcap_root, application, account, docname)
        if node:
            uri += '~~' + node
        return self._execute_request("DELETE", uri, username, domain, password, headers)


class PIDFManipulation(object):
    
    application_id = 'pidf-manipulation'
    
    def __init__(self, user, information):
        self.user = user
        self.information = information

    def to_xml(self):
        is_idle = ""
        if self.information.activity == "idle":        
            is_idle = "<rpid:user-input>idle</rpid:user-input>";
        
        body = """<?xml version="1.0" encoding="UTF-8"?>
                <presence xmlns='urn:ietf:params:xml:ns:pidf' xmlns:dm='urn:ietf:params:xml:ns:pidf:data-model' 
                  xmlns:rpid='urn:ietf:params:xml:ns:pidf:rpid' xmlns:c='urn:ietf:params:xml:ns:pidf:cipid' entity='sip:%(user)s'>
                    <tuple id='t9c4ce84e'>
                      <status>
                        <basic>open</basic>
                      </status>%(is_idle)s
                    </tuple>
                    <dm:person id='p00769964'>
                      <rpid:activities><rpid:%(activity)s/></rpid:activities>
                      <dm:note>%(note)s</dm:note>
                    </dm:person>
                  </presence>""" % {"user":  self.user,
                                    "is_idle": is_idle,
                                    "activity": self.information.activity,
                                    "note": self.information.note}
        return body


class PresenceRules(object):
    
    application_id = 'pres-rules'
    
    def __init__(self, user, policy):
        self.policy = policy

    def to_xml(self):
        allow = self.policy.allow
        deny = self.policy.deny
        confirm = self.policy.confirm
        ## build the allow rule
        transformations = """
        <transformations>
        <pr:provide-services>
        <pr:all-services/>
        </pr:provide-services>
        <pr:provide-persons>
        <pr:all-persons/>
        </pr:provide-persons>
        <pr:provide-devices>
        <pr:all-devices/>
        </pr:provide-devices>
        <pr:provide-all-attributes/>
        </transformations>"""        
        ##<?xml version='1.0' encoding='UTF-8'?>
        body = """<?xml version="1.0" encoding="UTF-8"?>
                  <ruleset xmlns="urn:ietf:params:xml:ns:common-policy" xmlns:pr="urn:ietf:params:xml:ns:pres-rules" xmlns:cp="www.counterpath.com/privacy-lists">"""
        if allow:
            rule = '<rule id="pres_whitelist"><conditions><identity>'
            for a in allow:
                if a.find('@') == -1:
                    ## a 'many' element
                    rule += '<many domain="%s"></many>' % a                    
                else:
                    ## a 'one' element
                    rule += '<one id="sip:%s"/>' % a
            rule += '</identity></conditions>'
            rule += '<actions><pr:sub-handling>allow</pr:sub-handling></actions>'
            rule += transformations
            rule += '</rule>'
            body += rule
        if deny:
            rule = '<rule id="pres_blacklist"><conditions><identity>'
            for a in deny:
                if a.find('@') == -1:
                    ## a 'many' element
                    rule += '<many domain="%s"></many>' % a                    
                else:
                    ## a 'one' element
                    rule += '<one id="sip:%s"/>' % a
            rule += '</identity></conditions>'
            rule += '<actions><pr:sub-handling>block</pr:sub-handling></actions>'
            rule += '<transformations/>'
            rule += '</rule>'
            body += rule
        if confirm:
            rule = '<rule id="pres_pending"><conditions><identity>'
            for a in confirm:
                if a.find('@') == -1:
                    ## a 'many' element
                    rule += '<many domain="%s"></many>' % a                    
                else:
                    ## a 'one' element
                    rule += '<one id="sip:%s"/>' % a
            rule += '</identity></conditions>'
            rule += '<actions><pr:sub-handling>confirm</pr:sub-handling></actions>'
            rule += '<transformations/>'
            rule += '</rule>'
            body += rule
        body += '</ruleset>'
        return body


if __name__ == '__main__':
    client = XCAPClient('http://secure.dns-hosting.info:433/xcap-root')
    #client.get_resource('31208005164@ag-projects.com', '5678', 'resource-lists')

    resource = PIDFManipulation("31208005164@ag-projects.com", "idle")

    #client.put_resource('31208005164@ag-projects.com', '5678', 'resource-lists', 'abcde')
    client.put_resource('31208005164@ag-projects.com', '5678', resource.application_id, resource.to_xml())
