'''SOAP engine for the Soap SIMPLE Proxy

Copyright 2007-2011 AG Projects
'''

from ssp.soap import tweaks
import SOAPpy
import sys, os, stat
import signal
import re
from time import time
from threading import RLock, currentThread
from StringIO import StringIO
from lxml import etree

from application.configuration import ConfigSection, ConfigSetting
from application import log
from application.process import process

from ssp import __version__
from ssp.soap.errors import *
from ssp.soap.types import *
from ssp.interfaces.xcap import XCAPClient, PIDFManipulation, PresenceRules

class ServiceURL(str):
    '''A HTTP URL with structure http[s]://address[:port]/path.'''
    def __new__(typ, value):
        if type(value) not in (str, unicode):
            raise TypeError, "%s should be a string" % typ.__name__
        pattern = r'^(http|https)://((([a-z0-9](([a-z0-9-]*?[a-z0-9])|)\.)*[a-z0-9](([a-z0-9-]*?[a-z0-9])|))|(\d{1,3}(\.\d{1,3}){3}))(:[1-9][0-9]{0,}|)(/[a-z0-9-_]+)*/?$'
        if not re.match(pattern, value, re.IGNORECASE):
            raise ValueError, "Invalid HTTP URL: '%s'." % (value,)
        if not value.endswith('/'): 
            value += '/'
        return str.__new__(typ, value)


class SOAPServerConfig(ConfigSection):
    __cfgfile__ = 'config.ini'
    __section__ = ('SOAP Server', 'WSDL')
    ## [SOAP Server] section
    host = '0.0.0.0'
    port = 9200
    log_requests = True
    dump_faults = True
    trace_http = False
    trace_soap = False
    debug = False
    threaded = True
    ## [WSDL] section
    wsdl = 'wsdl/ssp.wsdl'
    namespace = 'urn:AGProjects:SoapSIMPLEProxy'
    public_service_url = ServiceURL('https://example.com/soap-simple-proxy/')
    internal_service_url = ServiceURL('http://127.0.0.1:9200/')


class XCAPConfig(ConfigSection):
    __cfgfile__ = 'config.ini'
    __section__ = 'XCAP'
    root_uri = 'http://xcap.example.com/xcap-root'


class SOAPPorts: pass

## Initialize the SOAP engine (setup options)
SOAPpy.Config.authMethod       = '_authorize'
SOAPpy.Config.debug            = SOAPServerConfig.debug
SOAPpy.Config.dumpFaultInfo    = SOAPServerConfig.dump_faults
SOAPpy.Config.dumpHeadersIn    = SOAPServerConfig.trace_http
SOAPpy.Config.dumpHeadersOut   = SOAPServerConfig.trace_http
SOAPpy.Config.dumpSOAPIn       = SOAPServerConfig.trace_soap
SOAPpy.Config.dumpSOAPOut      = SOAPServerConfig.trace_soap
SOAPpy.Config.simplify_objects = 1 ## Do not change this!!! Will break the server!

class InterruptedError(Exception):
    pass

class SOAPServer:
    
    def __init__(self):
        self.__installSignalHandlers()
        self.server = self.__create()

    def __installSignalHandlers(self):
        process.signals.add_handler(signal.SIGHUP, self.__handle_SIGHUP)
        process.signals.add_handler(signal.SIGINT, self.__handle_SIGINT)
        process.signals.add_handler(signal.SIGTERM, self.__handle_SIGTERM)        

    def __create(self):
        address = (SOAPServerConfig.host, SOAPServerConfig.port)
        if SOAPServerConfig.threaded:
            server = SOAPpy.ThreadingSOAPServer(address, log=SOAPServerConfig.log_requests)
        else:
            server = SOAPpy.SOAPServer(address, log=SOAPServerConfig.log_requests)
        server.config.authMethod = '_authorize'
        namespace = SOAPServerConfig.namespace
        presence = PresencePort()
        self.xcap = XCAPClient(XCAPConfig.root_uri)
        SOAPPorts.presence = presence
        server.registerObject(presence, namespace + ':Presence')
        server.registerFunction(WSDLFile.wsdl)
        return server

    def __handle_SIGHUP(self, *args):
        log.msg("Received SIGHUP, shutting down.")
        self.stop()
        
    def __handle_SIGINT(self, *args):
        if process._daemon:
            log.msg("Received SIGINT, shutting down.")
        else:
            log.msg("Received KeyboardInterrupt, exiting.")
        self.stop()
        
    def __handle_SIGTERM(self, *args):
        log.msg("Received SIGTERM, shutting down.")
        self.stop()

    def run(self):
        try:
            self.server.serve_forever()
        except InterruptedError:
            pass
        
    def stop(self):
        raise InterruptedError()


class WSDLFile:
    _wsdl = {}
    lock = RLock()
    
    def wsdl_preprocess(wsdl):
        wsdl = wsdl.replace('#VERSION#', __version__)
        wsdl = wsdl.replace('#PUBLIC_PORT_LOCATION#', SOAPServerConfig.public_service_url)
        wsdl = wsdl.replace('#INTERNAL_PORT_LOCATION#', SOAPServerConfig.internal_service_url)
        return wsdl
    
    def wsdl(path):
        '''Todo: attach a timestamp to every file so that if it's newer re-read it'''
        if path == '/wsdl':
            path = SOAPServerConfig.wsdl
        WSDLFile.lock.acquire()
        try:
            try:
                return WSDLFile._wsdl[path]
            except KeyError:
                pass
            try:
                f = open(process.config_file(path))
            except Exception, e:
                print e
                WSDLFile._wsdl[path] = None
            else:
                try:
                    WSDLFile._wsdl[path] = WSDLFile.wsdl_preprocess(f.read())
                finally:
                    f.close()
            return WSDLFile._wsdl[path]
        finally:
            WSDLFile.lock.release()

    wsdl = staticmethod(wsdl)
    wsdl_preprocess = staticmethod(wsdl_preprocess)
    

##
## Returning SOAP data.
##
## When returning Structure like responses, we can use any of the following:
##
## 1. return the data (which is a python dictionary) directly
## 2. return DataType(data).SOAPout (where DataType is the Structure subclass
##    that corresponds to our data type - say CustomerAccount or SipAccount)
##
## Even though both cases work well, the second case returns better formatted
## SOAP messages which are easier to read and also are smaller by up to 30%
## For big structures, like CustomerAccount, using the second version is even
## a bit faster:
##   305 requests/sec for 2nd case and 260 requests/sec for the 1st case.
##
## On the other hand for small structures like SipId, there is no gain in size
## between the 2 cases, while returning data using the 1st alternative is 24%
## faster.
##
## However considering that most of our returned datatypes are big, using the 2nd
## alternative will be a gain from all points of view: readability, size and
## execution time
##

def strlimit(s, limit):
    assert limit >= 3, "minimum limit is 3"
    if len(s) <= limit:
        return s
    else:
        return s[:limit-3] + '...'

## The SOAP server interface
class SOAPPort(object):
    def _authorize(self, *args, **kwargs):
        if SOAPServerConfig.log_requests:
            # temporary hack. we should limit any string input
            # even when inside a dict/list/...
            try:
                body   = kwargs['_SOAPContext'].body
                method = body._keyord[0]
                namespace = body[method]._ns or SOAPServerConfig.namespace
                _args = SOAPpy.Types.simplify(body[method])
            except:
                raise SOAPError, "Invalid SOAP request"
            module = namespace.split(':')[-1]
            sip_id = _args["sipId"]
            #print "user requesting %s:%s(%s)" % (module, method, arguments)
            log.msg("%s:%s(%s@%s)" % (module, method, sip_id["username"], sip_id["domain"]))
        me = currentThread()
        return 1

    def _whoami(self):
        '''Return the name of the caller function'''
        return sys._getframe(1).f_code.co_name


class PresencePort(SOAPPort):
    _xcap = property(lambda self: XCAPClient(XCAPConfig.root_uri))

    def setPresenceInformation(self, *args, **kwargs):
        try:
            id = kwargs['sipId']
            password = kwargs['password']
            information = kwargs['information']
        except KeyError, why:
            raise MissingInputError, '%s is missing' % why[0]
        id = SipId(id)
        information = PresenceInformation(information)
        id = "%(username)s@%(domain)s" % id
        resource = PIDFManipulation(id, information)
        response = self._xcap.put_resource(id, password, resource.application_id, "index.xml", resource.to_xml())
        if str(response.code)[0] != "2":
            raise XCAPError(str(response.code))

    def getPresenceInformation(self, *args, **kwargs):
        try:
            id = kwargs['sipId']
            password = kwargs['password']
        except KeyError, why:
            raise MissingInputError, '%s is missing' % why[0]
        id = SipId(id)
        id = "%(username)s@%(domain)s" % id
        response = self._xcap.get_resource(id, password, 'pidf-manipulation', "index.xml")
        if response.code == 404:
            information = {'activity': '', 'note': ''}
            return PresenceInformation(information).SOAPout
        if str(response.code)[0] != "2":
            raise XCAPError(str(response.code))
        doc = response.body.strip()
        ## then get the allow/reject presence members from the xcap document
        xml_doc = etree.parse(StringIO(doc))
        ns_dict = {"dm": "urn:ietf:params:xml:ns:pidf:data-model", "rpid": "urn:ietf:params:xml:ns:pidf:rpid"}
        ## active watchers
        note = xml_doc.xpath("//dm:note", namespaces = ns_dict)
        if note[0].text is None:
            note = ''
        else:
            note = note[0].text.strip()
        activities = xml_doc.xpath('//rpid:activities', namespaces = ns_dict)
        activity =  activities[0].getchildren()[0].tag
        activity =  activity[activity.find('}') + 1:]
        information = {'activity': activity, 'note': note}
        return PresenceInformation(information).SOAPout

    def deletePresenceInformation(self, *args, **kwargs):
        try:
            id = kwargs['sipId']
            password = kwargs['password']
        except KeyError, why:
            raise MissingInputError, '%s is missing' % why[0]
        id = SipId(id)
        id = "%(username)s@%(domain)s" % id
        response = self._xcap.delete_resource(id, password, 'pidf-manipulation', "index.xml")
        if response.code != 404 and str(response.code)[0] != "2":
            raise XCAPError(str(response.code))

    def getWatchers(self, *args, **kwargs):
        try:
            id = kwargs['sipId']
            password = kwargs['password']
        except KeyError, why:
            raise MissingInputError, '%s is missing' % why[0]
        id = SipId(id)
        id = "%(username)s@%(domain)s" % id
        response = self._xcap.get_resource(id, password, "org.openxcap.watchers", "index.xml")
        if response.code == 404:
            return WatcherArray([]).SOAPout
        if str(response.code)[0] != "2":
            raise XCAPError(str(response.code))
        doc = response.body.strip()
        xml_doc = etree.parse(StringIO(doc))
        pattern = re.compile("^\{.*\}(.*)$") # removes the namespace, there must be a better way?
        watchers = [dict((pattern.match(element.tag).group(1), element.text) for element in watcher) for watcher in xml_doc.getroot()]
        for watcher in watchers:
            watcher["online"] = watcher["online"] == "true"
        return WatcherArray(watchers).SOAPout
    
    def setPolicy(self, *args, **kwargs):
        try:
            id = kwargs['sipId']
            password = kwargs['password']
            policy = kwargs['policy']
        except KeyError, why:
            raise MissingInputError, '%s is missing' % why[0]
        id = SipId(id)
        policy = PresencePolicy(policy)
        id = "%(username)s@%(domain)s" % id
        resource = PresenceRules(id, policy)
        response = self._xcap.put_resource(id, password, resource.application_id, "pres-rules", resource.to_xml())
        if str(response.code)[0] != "2":
            raise XCAPError(str(response.code))

    def getPolicy(self, *args, **kwargs):
        try:
            id = kwargs['sipId']
            password = kwargs['password']
        except KeyError, why:
            raise MissingInputError, '%s is missing' % why[0]
        id = SipId(id)
        id = "%(username)s@%(domain)s" % id
        response = self._xcap.get_resource(id, password, 'pres-rules', "pres-rules")
        if response.code == 404:
            policy = {'allow': [], 'deny': [], 'confirm': []}
            return PresencePolicy(policy).SOAPout            
        if str(response.code)[0] != "2":
            raise XCAPError(str(response.code))
        doc = response.body.strip()
        ## then get the allow/reject presence members from the xcap document
        xml_doc = etree.parse(StringIO(doc))
        ns_dict = {"pr": "urn:ietf:params:xml:ns:pres-rules","cp": "urn:ietf:params:xml:ns:common-policy"}
        ## active watchers
        one = xml_doc.xpath("/cp:ruleset/cp:rule[cp:actions/pr:sub-handling='allow']/cp:conditions/cp:identity/cp:one", namespaces = ns_dict)
        many = xml_doc.xpath("/cp:ruleset/cp:rule[cp:actions/pr:sub-handling='allow']/cp:conditions/cp:identity/cp:many", namespaces = ns_dict)
        active_watchers = [w.attrib['id'].split("sip:", 1)[1] for w in one]
        active_domains = [w.attrib['domain'] for w in many if w.get('domain')]
        allow = active_watchers
        allow.extend(active_domains)
        ## terminated watchers
        one = xml_doc.xpath("/cp:ruleset/cp:rule[cp:actions/pr:sub-handling='block']/cp:conditions/cp:identity/cp:one", namespaces = ns_dict)
        many = xml_doc.xpath("/cp:ruleset/cp:rule[cp:actions/pr:sub-handling='block']/cp:conditions/cp:identity/cp:many", namespaces = ns_dict)
        terminated_watchers = [w.attrib['id'].split("sip:", 1)[1] for w in one]
        terminated_domains = [w.attrib['domain'] for w in many if w.get('domain')]
        deny = terminated_watchers
        deny.extend(terminated_domains)
        
        one = xml_doc.xpath("/cp:ruleset/cp:rule[cp:actions/pr:sub-handling='confirm']/cp:conditions/cp:identity/cp:one", namespaces = ns_dict)
        many = xml_doc.xpath("/cp:ruleset/cp:rule[cp:actions/pr:sub-handling='confirm']/cp:conditions/cp:identity/cp:many", namespaces = ns_dict)
        pending_watchers = [w.attrib['id'].split("sip:", 1)[1] for w in one]
        pending_domains = [w.attrib['domain'] for w in many if w.get('domain')]
        confirm = pending_watchers
        confirm.extend(pending_domains)
        policy = {'allow': allow, 'deny': deny, 'confirm': confirm}
        return PresencePolicy(policy).SOAPout
